"""Module to help with the continuation or stopping of a table block.
"""

import copy
from typing import List, Optional, Tuple, cast

from pymarkdown.container_blocks.container_grab_bag import POGGER
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.leaf_blocks.table_block_tuple import TableTuple
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import TableBlockStackToken
from pymarkdown.tokens.table_markdown_tokens import (
    TableMarkdownBodyToken,
    TableMarkdownHeaderItemToken,
    TableMarkdownHeaderToken,
    TableMarkdownRowItemToken,
    TableMarkdownRowToken,
    TableMarkdownToken,
)
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

# pylint: disable=too-few-public-methods


class TableBlockContinuationHelper:
    """Class to help with the continuation or stopping of a table block."""

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def determine_continue_or_stop(
        parser_state: ParserState,
        position_marker: PositionMarker,
        was_started: bool,
        remaining_line_to_parse: str,
        extracted_whitespace: str,
        unmodified_line_to_parse: str,
        original_stack_depth: int,
        original_document_depth: int,
        end_table_index: int,
        line_to_parse_size: int,
        is_blank_line: bool,
        did_complete_table: bool,
        parsed_table_tuple: Optional[TableTuple],
        lines_to_requeue: List[str],
        process_mode: int,
    ) -> Tuple[bool, bool, List[MarkdownToken], bool]:
        """
        Determine whether to continue with the processing of the table.
        """
        # TODO leave in until testing of tables with tabs is complete.
        _ = (process_mode, extracted_whitespace)

        if did_pause_table := (
            end_table_index >= 0
            and end_table_index == line_to_parse_size
            and not is_blank_line
        ):
            POGGER.debug(">>parse_table>>continuation")
            TableBlockContinuationHelper.__add_line_for_table_continuation(
                parser_state,
                position_marker,
                was_started,
                remaining_line_to_parse,
                unmodified_line_to_parse,
                original_stack_depth,
                original_document_depth,
            )
        if not did_pause_table and was_started or did_complete_table:
            return TableBlockContinuationHelper.__stop_table_continuation(
                parser_state,
                did_complete_table,
                parsed_table_tuple,
                lines_to_requeue,
                did_pause_table,
            )

        POGGER.debug(">>parse_table>>other")
        return did_pause_table, False, [], False

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __stop_table_continuation(
        parser_state: ParserState,
        did_complete_table: bool,
        parsed_table_tuple: Optional[TableTuple],
        lines_to_requeue: List[str],
        did_pause_table: bool,
    ) -> Tuple[bool, bool, List[MarkdownToken], bool]:
        """
        As part of processing a table, stop a continuation.
        """
        POGGER.debug(">>parse_table>>no longer need start")
        if did_complete_table:
            assert parsed_table_tuple, "Table tuple must be defined by now."
            assert (
                parsed_table_tuple.normalized_destination is not None
            ), "normalized_destination must be defined by now."
            # TODO remove nd from parsed_table_tuple
            # did_add_definition = False  # LinkParseHelper.add_link_definition(                parsed_table_tuple.normalized_destination, parsed_table_tuple.link_titles            )
            table_stack_token = cast(TableBlockStackToken, parser_state.token_stack[-1])
            # assert (
            #     link_def_token.extracted_whitespace is not None
            # ), "extracted_whitespace must be defined by now."
            # extracted_whitespace = link_def_token.extracted_whitespace

            # POGGER.debug(
            #     "link_def_token.extracted_whitespace>:$:<",
            #     link_def_token.extracted_whitespace,
            # )
            # POGGER.debug(
            #     "link_def_token.continuation_lines>:$:<",
            #     link_def_token.continuation_lines,
            # )
            # POGGER.debug(
            #     "link_def_token.unmodified_lines>:$:<", link_def_token.unmodified_lines
            # )
            POGGER.debug("lines_to_requeue>:$:<", lines_to_requeue)

            does_any_line_have_tabs = any(
                ParserHelper.tab_character in ffg
                for ffg in table_stack_token.unmodified_lines
            )
            POGGER.debug("does_any_line_have_tabs>:$:<", does_any_line_have_tabs)

            # last_container_index = parser_state.find_last_container_on_stack()
            # if does_any_line_have_tabs and last_container_index > 0:
            #     (
            #         extracted_whitespace,
            #         parsed_table_tuple,
            #     ) = TableBlockContinuationHelper.__stop_table_continuation_with_tab(
            #         parser_state,
            #         table_stack_token,
            #         last_container_index,
            #         lines_to_requeue,
            #         process_mode,
            #         extracted_whitespace,
            #         parsed_table_tuple,
            #     )

            assert (
                parsed_table_tuple.normalized_destination is not None
            ), "normalized_destination must be defined by now."
            new_tokens: List[MarkdownToken] = []

            start_token = TableMarkdownToken(
                position_marker=table_stack_token.start_position_marker,
            )

            start_header_token = TableMarkdownHeaderToken(
                parsed_table_tuple.xyz[0],
                parsed_table_tuple.xyz[1],
                position_marker=table_stack_token.start_position_marker,
            )
            new_tokens.extend((start_token, start_header_token))

            for next_column_index, next_column in enumerate(
                parsed_table_tuple.xyz[0].columns
            ):
                start_header_item_token = TableMarkdownHeaderItemToken(
                    next_column.leading_whitespace,
                    parsed_table_tuple.col_as[next_column_index],
                    position_marker=table_stack_token.start_position_marker,
                )
                new_tokens.extend(
                    (
                        start_header_item_token,
                        TextMarkdownToken(
                            next_column.text,
                            "",
                            position_marker=table_stack_token.start_position_marker,
                        ),
                        start_header_item_token.generate_close_markdown_token_from_markdown_token(
                            next_column.trailing_whitespace, ""
                        ),
                    )
                )
            new_tokens.append(
                start_header_token.generate_close_markdown_token_from_markdown_token(
                    "", ""
                )
            )

            if len(parsed_table_tuple.xyz) > 2:
                TableBlockContinuationHelper.__stop_table_continuation_body(
                    new_tokens, table_stack_token, parsed_table_tuple
                )

            new_tokens.append(
                start_token.generate_close_markdown_token_from_markdown_token("", "")
            )

            # POGGER.debug(">>link_info>>$", parsed_table_tuple.link_info)
            # assert parsed_table_tuple.link_info.line_destination_whitespace is not None
            # POGGER.debug(
            #     ">>line_destination_whitespace>>$",
            #     ParserHelper.make_whitespace_visible(
            #         parsed_table_tuple.link_info.line_destination_whitespace
            #     ),
            # )
            TableBlockContinuationHelper.__stop_table_continuation_end(
                parser_state, new_tokens
            )
            return did_pause_table, len(lines_to_requeue) > 1, new_tokens, False

        del parser_state.token_stack[-1]
        return did_pause_table, True, [], True

    @staticmethod
    def __stop_table_continuation_end(
        parser_state: ParserState, new_tokens: List[MarkdownToken]
    ) -> None:
        POGGER.debug(">>new_tokens>>$", new_tokens)
        del parser_state.token_stack[-1]
        if parser_state.token_stack[-1].is_paragraph:
            tokens_from_close, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=(len(parser_state.token_stack) - 1),
            )
            assert len(tokens_from_close) == 1, "Only one token should be returned."
            new_tokens.insert(0, tokens_from_close[0])

    # pylint: disable=too-many-locals
    @staticmethod
    def __stop_table_continuation_body(
        new_tokens: List[MarkdownToken],
        table_stack_token: TableBlockStackToken,
        parsed_table_tuple: TableTuple,
    ) -> None:
        start_body_token = TableMarkdownBodyToken(
            position_marker=table_stack_token.start_position_marker
        )
        new_tokens.append(start_body_token)

        for next_row_index in range(2, len(parsed_table_tuple.xyz)):
            x = parsed_table_tuple.xyz[next_row_index]

            abc = x.columns[: len(parsed_table_tuple.col_as)]
            if abc_after := x.columns[len(parsed_table_tuple.col_as) :]:
                aaa: List[str] = []
                for ii in abc_after:
                    aaa.extend((ii.leading_whitespace, ii.text, ii.trailing_whitespace))
                aaa_string = "".join(aaa)
            else:
                aaa_string = ""
            delta = len(parsed_table_tuple.col_as) - len(abc)

            start_row_token = TableMarkdownRowToken(
                x.extracted_whitespace,
                x.trailing_whitespace,
                x.did_start_with_separator,
                delta,
                position_marker=table_stack_token.start_position_marker,
            )
            new_tokens.append(start_row_token)

            for next_column_index, next_column in enumerate(abc):
                start_row_item_token = TableMarkdownRowItemToken(
                    next_column.leading_whitespace,
                    parsed_table_tuple.col_as[next_column_index],
                    position_marker=table_stack_token.start_position_marker,
                )

                new_tokens.extend(
                    (
                        start_row_item_token,
                        TextMarkdownToken(
                            next_column.text,
                            "",
                            position_marker=table_stack_token.start_position_marker,
                        ),
                        start_row_item_token.generate_close_markdown_token_from_markdown_token(
                            next_column.trailing_whitespace, ""
                        ),
                    )
                )
            new_tokens.append(
                start_row_token.generate_close_markdown_token_from_markdown_token(
                    aaa_string, ""
                )
            )
        new_tokens.append(
            start_body_token.generate_close_markdown_token_from_markdown_token("", "")
        )

    # pylint: enable=too-many-locals

    # @staticmethod
    # def __stop_table_continuation_with_tab(
    #     parser_state: ParserState,
    #     table_stack_token: TableBlockStackToken,
    #     last_container_index: int,
    #     lines_to_requeue: List[str],
    #     process_mode: int,
    #     extracted_whitespace: str,
    #     parsed_table_tuple: TableTuple,
    # ) -> Tuple[str, TableTuple]:
    #     assert False
    #     POGGER.debug(
    #         "extracted_whitespace>:$:<",
    #         table_stack_token.extracted_whitespace,
    #     )
    #     assert parser_state.token_stack[
    #         last_container_index
    #     ].is_block_quote, "Container must be a block quote."
    #     last_block_quote_index = parser_state.find_last_block_quote_on_stack()
    #     last_block_quote_token = parser_state.token_stack[last_block_quote_index]
    #     block_quote_token = cast(
    #         BlockQuoteMarkdownToken, last_block_quote_token.matching_markdown_token
    #     )

    #     POGGER.debug(
    #         "table_stack_token.continuation_lines>:$:<", table_stack_token.continuation_lines
    #     )
    #     POGGER.debug(
    #         "table_stack_token.unmodified_lines>:$:<", table_stack_token.unmodified_lines
    #     )

    #     if len(table_stack_token.continuation_lines) == 1:
    #         extracted_whitespace = TableBlockContinuationHelper.__stop_table_continuation_with_tab_single(
    #             parser_state,
    #             table_stack_token,
    #             process_mode,
    #             block_quote_token,
    #             lines_to_requeue,
    #         )
    #     else:
    #         (
    #             extracted_whitespace,
    #             parsed_table_tuple,
    #         ) = TableBlockContinuationHelper.__stop_table_continuation_with_tab_multiple(
    #             parser_state, extracted_whitespace, table_stack_token, block_quote_token
    #         )

    #     return extracted_whitespace, parsed_table_tuple

    # @staticmethod
    # def __stop_table_continuation_with_tab_single(
    #     parser_state: ParserState,
    #     table_stack_token: TableBlockStackToken,
    #     process_mode: int,
    #     block_quote_token: BlockQuoteMarkdownToken,
    #     lines_to_requeue: List[str],
    # ) -> str:
    #     parsed_lines = table_stack_token.continuation_lines[0]
    #     original_lines = table_stack_token.unmodified_lines[0]

    #     (
    #         extracted_whitespace,
    #         split_tab,
    #         _,
    #     ) = TableBlockContinuationHelper.__find_line_ws(
    #         parsed_lines, original_lines
    #     )

    #     POGGER.debug("process_mode>:$:<", process_mode)
    #     POGGER.debug(
    #         "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
    #     )
    #     if process_mode == 1:
    #         block_quote_token.remove_last_bleading_space()
    #     else:
    #         for _ in lines_to_requeue:
    #             block_quote_token.remove_last_bleading_space()

    #     if split_tab:
    #         TabHelper.adjust_block_quote_indent_for_tab(parser_state)
    #     POGGER.debug(
    #         "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
    #     )
    #     return extracted_whitespace

    # @staticmethod
    # def __stop_table_continuation_with_tab_multiple(
    #     parser_state: ParserState,
    #     extracted_whitespace: str,
    #     table_stack_token: TableBlockStackToken,
    #     block_quote_token: BlockQuoteMarkdownToken,
    # ) -> Tuple[str, TableTuple]:
    #     assert False
    #     split_tabs_list: List[bool] = []
    #     completed_table_text: str = ""
    #     alt_ws: Optional[str] = None
    #     for this_line_index, this_line in enumerate(table_stack_token.continuation_lines):
    #         (
    #             completed_table_text,
    #             extracted_whitespace,
    #             alt_ws,
    #         ) = TableBlockContinuationHelper.__stop_table_continuation_with_tab_multiple_loop(
    #             table_stack_token,
    #             this_line_index,
    #             this_line,
    #             completed_table_text,
    #             extracted_whitespace,
    #             alt_ws,
    #             split_tabs_list,
    #         )

    #     POGGER.debug("completed_table_text>:$:<", completed_table_text)
    #     assert alt_ws is not None, "This value must be set inside of the for loop."

    #     (
    #         did_succeed,
    #         next_index,
    #         new_parsed_table_tuple,
    #     ) = TableParseHelper.parse_table(
    #         parser_state, completed_table_text, 0, alt_ws, True, ""
    #     )
    #     assert (
    #         did_succeed
    #     ), "Since this is the stop and there is at least one valid match, this must be true."
    #     assert (
    #         len(completed_table_text) == next_index
    #     ), "Index must be at the end of the stirng."
    #     assert new_parsed_table_tuple is not None, "New tuple must be defined."

    #     TableBlockContinuationHelper.__xx_multiple_fix_leading_spaces(
    #         block_quote_token, split_tabs_list, table_stack_token
    #     )
    #     return extracted_whitespace, new_parsed_table_tuple

    # # pylint: disable=too-many-arguments
    # @staticmethod
    # def __stop_table_continuation_with_tab_multiple_loop(
    #     table_stack_token: TableBlockStackToken,
    #     this_line_index: int,
    #     this_line: str,
    #     completed_table_text: str,
    #     extracted_whitespace: str,
    #     alt_ws: Optional[str],
    #     split_tabs_list: List[bool],
    # ) -> Tuple[str, str, Optional[str]]:
    #     assert False
    #     original_this_line = table_stack_token.unmodified_lines[this_line_index]
    #     POGGER.debug("this_line_index>:$:<", this_line_index)
    #     POGGER.debug("this_line>:$:<", this_line)
    #     POGGER.debug("original_this_line>:$:<", original_this_line)

    #     (
    #         extracted_ws,
    #         split_tab,
    #         start_whitespace_index,
    #     ) = TableBlockContinuationHelper.__find_line_ws(
    #         this_line, original_this_line
    #     )

    #     if completed_table_text:
    #         completed_table_text += "\n"
    #     if this_line_index == 0:
    #         extracted_whitespace = extracted_ws
    #         alt_ws = TabHelper.detabify_string(
    #             extracted_whitespace, start_whitespace_index
    #         )
    #     else:
    #         completed_table_text += extracted_ws
    #     completed_table_text += this_line
    #     split_tabs_list.append(split_tab)
    #     return completed_table_text, extracted_whitespace, alt_ws

    # # pylint: enable=too-many-arguments

    # @staticmethod
    # def __find_line_ws(parsed_lines: str, original_lines: str) -> Tuple[str, bool, int]:
    #     start_text_index = original_lines.find(parsed_lines)
    #     assert start_text_index != -1, "Index must be found within string."
    #     POGGER.debug("start_text_index>:$:<", start_text_index)
    #     start_whitespace_index, _ = ParserHelper.extract_spaces_from_end(
    #         original_lines, start_text_index
    #     )
    #     POGGER.debug("start_whitespace_index>:$:<", start_whitespace_index)
    #     tabified_whitespace = original_lines[start_whitespace_index:start_text_index]
    #     POGGER.debug("tabified_whitespace>:$:<", tabified_whitespace)
    #     split_tab = bool(tabified_whitespace and tabified_whitespace[0] == "\t")
    #     if not split_tab:
    #         tabified_whitespace = tabified_whitespace[1:]
    #         start_whitespace_index += 1
    #     extracted_whitespace = tabified_whitespace
    #     POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
    #     POGGER.debug("split_tab>:$:<", split_tab)
    #     return extracted_whitespace, split_tab, start_whitespace_index

    # @staticmethod
    # def __xx_multiple_fix_leading_spaces(
    #     block_quote_token: BlockQuoteMarkdownToken,
    #     split_tabs_list: List[bool],
    #     table_stack_token: TableBlockStackToken,
    # ) -> None:
    #     POGGER.debug("split_tabs_list>:$:<", split_tabs_list)
    #     POGGER.debug(
    #         "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
    #     )
    #     assert (
    #         block_quote_token.bleading_spaces is not None
    #     ), "Bleading spaces must be defined by now."
    #     leading_spaces: List[str] = []
    #     for _ in table_stack_token.continuation_lines:
    #         last_leading_space = block_quote_token.remove_last_bleading_space()
    #         POGGER.debug("last_leading_space>:$:<", last_leading_space)
    #         # if last_leading_space[0] == "\n":
    #         #     last_leading_space = last_leading_space[1:]
    #         leading_spaces.append(last_leading_space)
    #     assert len(split_tabs_list) == len(
    #         leading_spaces
    #     ), "The two lists must have the same length."
    #     POGGER.debug("leading_spaces>:$:<", leading_spaces)
    #     POGGER.debug(
    #         "block_quote_token.leading_spaces>:$:<", block_quote_token.bleading_spaces
    #     )
    #     is_first = not block_quote_token.bleading_spaces
    #     for prefix_to_add in leading_spaces:
    #         if split_tabs_list[0]:
    #             prefix_to_add = prefix_to_add[:-1]
    #         del split_tabs_list[0]
    #         POGGER.debug(
    #             "__xx_multiple_fix_leading_spaces>>block_token>>$", block_quote_token
    #         )
    #         block_quote_token.add_bleading_spaces(prefix_to_add, is_first)
    #         POGGER.debug(
    #             "__xx_multiple_fix_leading_spaces>>block_token>>$", block_quote_token
    #         )
    #         is_first = False

    # pylint: disable=too-many-arguments
    @staticmethod
    def __add_line_for_table_continuation(
        parser_state: ParserState,
        position_marker: PositionMarker,
        was_started: bool,
        remaining_line_to_parse: str,
        unmodified_line_to_parse: str,
        original_stack_depth: int,
        original_document_depth: int,
    ) -> None:
        """
        As part of processing a table, add a line to the continuation.
        """

        line_to_store = remaining_line_to_parse
        if not was_started:
            POGGER.debug(">>parse_table>>marking start")
            new_token = TableBlockStackToken(position_marker)
            parser_state.token_stack.append(new_token)
            new_token.original_stack_depth = original_stack_depth
            new_token.original_document_depth = original_document_depth

            new_token.last_block_quote_stack_token = (
                parser_state.last_block_quote_stack_token
            )
            new_token.last_block_quote_markdown_token_index = (
                parser_state.last_block_quote_markdown_token_index
            )
            new_token.copy_of_last_block_quote_markdown_token = (
                parser_state.copy_of_last_block_quote_markdown_token
            )

            lbqi = parser_state.find_last_block_quote_on_stack()
            new_token.last_block_quote_stack_token = (
                parser_state.token_stack[lbqi] if lbqi else None
            )
            new_token.last_block_quote_markdown_token_index = None
            new_token.copy_of_last_block_quote_markdown_token = None
            if new_token.last_block_quote_stack_token:
                markdown_token = parser_state.token_stack[lbqi].matching_markdown_token
                assert markdown_token is not None
                new_token.last_block_quote_markdown_token_index = (
                    parser_state.token_document.index(markdown_token)
                )
                new_token.copy_of_last_block_quote_markdown_token = cast(
                    BlockQuoteMarkdownToken,
                    copy.deepcopy(
                        parser_state.token_document[
                            new_token.last_block_quote_markdown_token_index
                        ]
                    ),
                )

            new_token.copy_of_token_stack = parser_state.copy_of_token_stack

            new_token.x1_token = parser_state.x1_token
            new_token.copy_of_x1_token = parser_state.copy_of_x1_token
            new_token.x1_token_index = parser_state.x1_token_index
        else:
            new_token = cast(TableBlockStackToken, parser_state.token_stack[-1])

        POGGER.debug(">>line_to_store>>add>:$<<", line_to_store)
        POGGER.debug(">>unmodified_line_to_parse>>add>:$<<", unmodified_line_to_parse)
        assert unmodified_line_to_parse.endswith(
            line_to_store
        ), "Unmodified line must end with the processed line."
        new_token.add_continuation_line(line_to_store)
        new_token.add_unmodified_line(unmodified_line_to_parse)

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to test that the return code schemes are set up properly.
"""

import os
import runpy
from test.markdown_scanner import MarkdownScanner
from test.utils import (
    assert_file_is_as_expected,
    copy_to_temp_file,
    create_temporary_configuration_file,
)


def test_markdown_return_code_command_line_bad():
    """
    Test to...
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["--return-code-scheme", "invalid", "scan"]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py [-h] [-e ENABLE_RULES] [-d DISABLE_RULES]
               [--add-plugin ADD_PLUGIN] [--config CONFIGURATION_FILE]
               [--set SET_CONFIGURATION] [--strict-config] [--stack-trace]
               [--continue-on-error]
               [--log-level {CRITICAL,ERROR,WARNING,INFO,DEBUG}]
               [--log-file LOG_FILE] [--return-code-scheme {default,minimal}]
               {extensions,fix,plugins,scan,scan-stdin,version} ...
main.py: error: argument --return-code-scheme: invalid __validate_return_code_scheme value: 'invalid'"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_default_success():
    """
    Test to make sure a return code of 0 for SUCCESS is a default.

    This function shadows
    test_markdown_with_version
    and is shadowed by
    test_markdown_return_code_minimal_success
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["version"]

    version_path = os.path.join(".", "pymarkdown", "version.py")
    version_meta = runpy.run_path(version_path)
    semantic_version = version_meta["__version__"]

    expected_return_code = 0
    expected_output = """{version}
""".replace(
        "{version}", semantic_version
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_default_no_files_to_scan(caplog):
    """
    Test to make sure a return code of 1 for NO_FILES_TO_SCAN is a default.

    This function shadows
    test_markdown_with_direct_args
    and is shadowed by
    test_markdown_return_code_minimal_no_files_to_scan
    """

    # Arrange
    scanner = MarkdownScanner(use_main=False)
    supplied_arguments = ["--log-level", "DEBUG", "scan", "does-not-exist.md"]

    expected_return_code = 1
    expected_output = ""
    expected_error = """Provided path 'does-not-exist.md' does not exist.


No matching files found."""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, use_direct_arguments=True
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    assert "Using direct arguments: [" in caplog.text


def test_markdown_return_code_default_command_line_error():
    """
    Test to make sure a return code of 2 for COMMAND_LINE_ERROR is a default.

    This function shadows
    test_markdown_with_dash_l_only
    and is shadowed by
    test_markdown_return_code_minimal_command_line_error
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["scan", "-l"]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: the following arguments are required: path
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_default_fixed_at_least_one_file():
    """
    Test to make sure a return code of 3 for FIXED_AT_LEAST_ONE_FILE is a default.

    This function shadows
    test_md009 fix good_unordered_list_item_empty_lines_with_config_strict without the list_item_empty_lines config
    and is shadowed by
    test_markdown_return_code_minimal_fixed_at_least_one_file
    """

    # Arrange
    scanner = MarkdownScanner()
    source_file_contents = """- list item text
\a\a
  list item text
""".replace(
        "\a", " "
    )
    with create_temporary_configuration_file(
        source_file_contents, file_name_suffix=".md"
    ) as temp_source_path:
        supplied_arguments = [
            "--set",
            "plugins.md009.strict=$!True",
            "--strict-config",
            "fix",
            temp_source_path,
        ]

        expected_return_code = 3
        expected_output = f"Fixed: {temp_source_path}"
        expected_error = ""
        expected_file_contents = """- list item text

  list item text
"""

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
        assert_file_is_as_expected(temp_source_path, expected_file_contents)


def test_markdown_return_code_default_scan_triggered_at_least_once():
    """
    Test to make sure a return code of 1 for SCAN_TRIGGERED_AT_LEAST_ONCE is a default.

    This function shadows
    test_md009_good_unordered_list_item_empty_lines_with_config_strict
    and is shadowed by
    test_markdown_return_code_minimal_scan_triggered_at_least_once
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md009", "good_unordered_list_item_empty_lines.md"
    )
    supplied_arguments = [
        "--set",
        "plugins.md009.strict=$!True",
        "--strict-config",
        "scan",
        source_path,
    ]

    expected_return_code = 1
    expected_output = (
        f"{source_path}:2:1: "
        + "MD009: Trailing spaces "
        + "[Expected: 0; Actual: 2] (no-trailing-spaces)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_default_system_error():
    """
    Test to make sure a return code of 1 for SYSTEM_ERROR is a default.

    This function shadows
    test_markdown_with_dash_dash_add_plugin_and_bad_path
    and is shadowed by
    test_markdown_return_code_minimal_system_error
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "--add-plugin",
        "MD047",
        "scan",
        source_path,
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while loading plugins:\n"
        + "Plugin path 'MD047' does not exist.\n"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_minimal_success():
    """
    Test to make sure a return code of 0 for SUCCESS is applied with the minimal scheme.

    This function shadows
    test_markdown_return_code_default_success
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["--return-code-scheme", "minimal", "version"]

    version_path = os.path.join(".", "pymarkdown", "version.py")
    version_meta = runpy.run_path(version_path)
    semantic_version = version_meta["__version__"]

    expected_return_code = 0
    expected_output = """{version}
""".replace(
        "{version}", semantic_version
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_minimal_no_files_to_scan(caplog):
    """
    Test to make sure a return code of 0 for NO_FILES_TO_SCAN is applied with the minimal scheme.

    This function shadows
    test_markdown_return_code_default_no_files_to_scan
    """

    # Arrange
    scanner = MarkdownScanner(use_main=False)
    supplied_arguments = [
        "--log-level",
        "DEBUG",
        "--return-code-scheme",
        "minimal",
        "scan",
        "does-not-exist.md",
    ]

    expected_return_code = 0
    expected_output = ""
    expected_error = """Provided path 'does-not-exist.md' does not exist.


No matching files found."""

    # Act
    execute_results = scanner.invoke_main(
        arguments=supplied_arguments, use_direct_arguments=True
    )

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )

    assert "Using direct arguments: [" in caplog.text


def test_markdown_return_code_minimal_command_line_error():
    """
    Test to make sure a return code of 2 for COMMAND_LINE_ERROR is applied with the minimal scheme.

    This function shadows
    test_markdown_return_code_default_command_line_error
    """

    # Arrange
    scanner = MarkdownScanner()
    supplied_arguments = ["--return-code-scheme", "minimal", "scan", "-l"]

    expected_return_code = 2
    expected_output = ""
    expected_error = """usage: main.py scan [-h] [-l] [-r] [-ae ALTERNATE_EXTENSIONS]
                    [-e PATH_EXCLUSIONS]
                    path [path ...]
main.py scan: error: the following arguments are required: path
"""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_minimal_fixed_at_least_one_file():
    """
    Test to make sure a return code of 0 for FIXED_AT_LEAST_ONE_FILE is applied with the minimal scheme.

    This function shadows
    test_markdown_return_code_default_fixed_at_least_one_file
    """

    # Arrange
    scanner = MarkdownScanner()
    with copy_to_temp_file(
        os.path.join(
            "test",
            "resources",
            "rules",
            "md009",
            "good_unordered_list_item_empty_lines.md",
        )
    ) as temp_source_path:
        supplied_arguments = [
            "--set",
            "plugins.md009.strict=$!True",
            "--set",
            "plugins.md009.list_item_empty_lines=$!True",
            "--strict-config",
            "--return-code-scheme",
            "minimal",
            "fix",
            temp_source_path,
        ]

        expected_return_code = 0
        expected_output = f"Fixed: {temp_source_path}"
        expected_error = ""
        # expected_file_contents = generate_md009_expected_contents(temp_source_path, 2)

        # Act
        execute_results = scanner.invoke_main(arguments=supplied_arguments)

        # Assert
        execute_results.assert_results(
            expected_output, expected_error, expected_return_code
        )
        # assert_file_is_as_expected(temp_source_path, expected_file_contents)


def test_markdown_return_code_minimal_scan_triggered_at_least_once():
    """
    Test to make sure a return code of 0 for SCAN_TRIGGERED_AT_LEAST_ONCE is applied with the minimal scheme.

    This function shadows
    test_markdown_return_code_default_scan_triggered_at_least_once
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md009", "good_unordered_list_item_empty_lines.md"
    )
    supplied_arguments = [
        "--set",
        "plugins.md009.strict=$!True",
        "--strict-config",
        "--return-code-scheme",
        "minimal",
        "scan",
        source_path,
    ]

    expected_return_code = 0
    expected_output = (
        f"{source_path}:2:1: "
        + "MD009: Trailing spaces "
        + "[Expected: 0; Actual: 2] (no-trailing-spaces)"
    )
    expected_error = ""

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )


def test_markdown_return_code_minimal_system_error():
    """
    Test to make sure a return code of 1 for SYSTEM_ERROR is applied with the minimal scheme.

    This function shadows
    test_markdown_return_code_default_system_error
    """

    # Arrange
    scanner = MarkdownScanner()
    source_path = os.path.join(
        "test", "resources", "rules", "md047", "end_with_blank_line.md"
    )
    supplied_arguments = [
        "--add-plugin",
        "MD047",
        "--return-code-scheme",
        "minimal",
        "scan",
        source_path,
    ]

    expected_return_code = 1
    expected_output = ""
    expected_error = (
        "BadPluginError encountered while loading plugins:\n"
        + "Plugin path 'MD047' does not exist.\n"
    )

    # Act
    execute_results = scanner.invoke_main(arguments=supplied_arguments)

    # Assert
    execute_results.assert_results(
        expected_output, expected_error, expected_return_code
    )



"""
Setup file for the PyMarkdown project.
"""

import runpy

from setuptools import setup


def parse_requirements():
    with open("install-requirements.txt", "r", encoding="utf-8") as requirement_file:
        lineiter = [line.strip() for line in requirement_file]
    return [line for line in lineiter if line and not line.startswith("#")]


def get_semantic_version():
    version_meta = runpy.run_path("./pymarkdown/version.py")
    return version_meta["__version__"]


def load_readme_file():
    with open("README.md", "r", encoding="utf-8") as readme_file:
        return readme_file.read()


AUTHOR = "Jack De Winter"
AUTHOR_EMAIL = "jack.de.winter@outlook.com"
PROJECT_URL = "https://github.com/jackdewinter/pymarkdown"
PROJECT_URLS = {
    "Change Log": "https://pymarkdown.readthedocs.io/en/latest/changelog/",
}

PACKAGE_NAME = "pymarkdownlnt"
SEMANTIC_VERSION = get_semantic_version()
MINIMUM_PYTHON_VERSION = "3.9.0"

ONE_LINE_DESCRIPTION = "A GitHub Flavored Markdown compliant Markdown linter."
LONG_DESCRIPTION = load_readme_file()
LONG_DESCRIPTION_CONTENT_TYPE = "text/markdown"

KEYWORDS = ["markdown", "linter", "markdown linter"]
PROJECT_CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Natural Language :: English",
]

PACKAGE_MODULES = [
    "pymarkdown",
    "pymarkdown.block_quotes",
    "pymarkdown.coalesce",
    "pymarkdown.container_blocks",
    "pymarkdown.extension_manager",
    "pymarkdown.extensions",
    "pymarkdown.general",
    "pymarkdown.html",
    "pymarkdown.inline",
    "pymarkdown.leaf_blocks",
    "pymarkdown.links",
    "pymarkdown.list_blocks",
    "pymarkdown.plugin_manager",
    "pymarkdown.plugins",
    "pymarkdown.plugins.utils",
    "pymarkdown.resources",
    "pymarkdown.tokens",
    "pymarkdown.transform_gfm",
    "pymarkdown.transform_markdown",
]

setup(
    name=PACKAGE_NAME,
    version=SEMANTIC_VERSION,
    python_requires=f">={MINIMUM_PYTHON_VERSION}",
    install_requires=parse_requirements(),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR,
    maintainer_email=AUTHOR_EMAIL,
    url=PROJECT_URL,
    description=ONE_LINE_DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    keywords=KEYWORDS,
    classifiers=PROJECT_CLASSIFIERS,
    project_urls=PROJECT_URLS,
    entry_points={
        "console_scripts": [
            "pymarkdown=pymarkdown.__main__:main",
            "pymarkdownlnt=pymarkdown.__main__:main",
        ],
    },
    packages=PACKAGE_MODULES,
    include_package_data=True,
)



"""
Module to handle the processing of configuration for the application.
"""

import argparse
import json
import logging
import os
from typing import Callable, Optional

import tomli
import yaml
from application_properties import (
    ApplicationProperties,
    ApplicationPropertiesJsonLoader,
    ApplicationPropertiesTomlLoader,
    ApplicationPropertiesUtilities,
    ApplicationPropertiesYamlLoader,
)

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class ApplicationConfigurationHelper:
    """
    Class to handle the processing of configuration for the application.
    """

    __default_configuration_file = ".pymarkdown"
    __default_yaml_configuration_file_extension = ".yaml"
    __other_default_yaml_configuration_file_extension = ".yml"

    @staticmethod
    def apply_configuration_layers(
        args: argparse.Namespace,
        properties: ApplicationProperties,
        handle_error: Callable[[str, Optional[Exception]], None],
    ) -> None:
        """
        Apply any general python configuration files followed by any configuration
        files specific to this project.
        """
        LOGGER.debug("Looking for any standard python configuration files.")
        ApplicationPropertiesUtilities.process_standard_python_configuration_files(
            properties, handle_error
        )

        LOGGER.debug("Looking for application specific configuration files.")
        ApplicationConfigurationHelper.__process_project_specific_json_configuration(
            args,
            properties,
            handle_error,
        )

        if args.strict_configuration or properties.get_boolean_property(
            "mode.strict-config", strict_mode=True
        ):
            properties.enable_strict_mode()

    @staticmethod
    def __process_project_specific_json_configuration(
        args: argparse.Namespace,
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> None:
        """
        Load configuration information from JSON configuration files.
        """

        # Look for the default configuration files in the current working directory.
        ApplicationConfigurationHelper.__process_default_configuration_files(
            application_properties, handle_error_fn
        )

        # A configuration file specified on the command line has a higher precedence
        # than anything except a specific setting applied on the command line.
        if args.configuration_file:
            if not os.path.isfile(args.configuration_file):
                handle_error_fn(
                    f"Specified configuration file `{args.configuration_file}` does not exist.",
                    None,
                )

            LOGGER.debug(
                "Determining file type for specified configuration file '%s'.",
                args.configuration_file,
            )
            try:
                with open(args.configuration_file, encoding="utf-8") as infile:
                    json.load(infile)
                did_load_as_json = True
            except json.decoder.JSONDecodeError:
                did_load_as_json = False

            try:
                with open(args.configuration_file, "rb") as infile:
                    loaded_document = yaml.safe_load(infile)
                did_load_as_yaml = not isinstance(loaded_document, str)
            except yaml.MarkedYAMLError:
                did_load_as_yaml = False

            try:
                with open(args.configuration_file, "rb") as infile:
                    tomli.load(infile)
                did_load_as_toml = True
            except tomli.TOMLDecodeError:
                did_load_as_toml = False

            if did_load_as_json:
                LOGGER.debug(
                    "Attempting to load configuration file '%s' as a JSON file.",
                    args.configuration_file,
                )
                ApplicationPropertiesJsonLoader.load_and_set(
                    application_properties,
                    args.configuration_file,
                    handle_error_fn=handle_error_fn,
                    clear_property_map=False,
                    check_for_file_presence=False,
                )
            elif did_load_as_yaml:
                LOGGER.debug(
                    "Attempting to load configuration file '%s' as a YAML file.",
                    args.configuration_file,
                )
                ApplicationPropertiesYamlLoader.load_and_set(
                    application_properties,
                    args.configuration_file,
                    handle_error_fn=handle_error_fn,
                    clear_property_map=False,
                    check_for_file_presence=False,
                )
            elif did_load_as_toml:
                LOGGER.debug(
                    "Attempting to load configuration file '%s' as a TOML file.",
                    args.configuration_file,
                )
                ApplicationPropertiesTomlLoader.load_and_set(
                    application_properties,
                    args.configuration_file,
                    handle_error_fn=handle_error_fn,
                    clear_property_map=False,
                    check_for_file_presence=False,
                )
            else:
                formatted_error = f"Specified configuration file '{args.configuration_file}' was not parseable as a JSON, YAML, or TOML file."
                LOGGER.warning(formatted_error)
                handle_error_fn(formatted_error, None)

        # A specific setting applied on the command line has the highest precedence.
        if args.set_configuration:
            LOGGER.debug(
                "Attempting to set one or more provided manual properties '%s'.",
                args.set_configuration,
            )
            application_properties.set_manual_property(args.set_configuration)

    @staticmethod
    def __process_default_configuration_files(
        application_properties: ApplicationProperties,
        handle_error_fn: Callable[[str, Optional[Exception]], None],
    ) -> None:
        abs_file_name = os.path.abspath(
            ApplicationConfigurationHelper.__default_configuration_file
        )
        LOGGER.debug(
            "Attempting to find/load '%s' as a default JSON configuration file.",
            abs_file_name,
        )
        (
            did_apply_map,
            did_have_one_error,
        ) = ApplicationPropertiesJsonLoader.load_and_set(
            application_properties,
            abs_file_name,
            handle_error_fn=handle_error_fn,
            clear_property_map=False,
            check_for_file_presence=True,
        )
        if not did_apply_map and not did_have_one_error:
            new_file_name = (
                abs_file_name
                + ApplicationConfigurationHelper.__default_yaml_configuration_file_extension
            )
            LOGGER.debug(
                "Attempting to find/load '%s' as a default YAML configuration file.",
                new_file_name,
            )
            (
                did_apply_map,
                did_have_one_error,
            ) = ApplicationPropertiesYamlLoader.load_and_set(
                application_properties,
                new_file_name,
                handle_error_fn=handle_error_fn,
                clear_property_map=False,
                check_for_file_presence=True,
            )
        if not did_apply_map and not did_have_one_error:
            new_file_name = (
                abs_file_name
                + ApplicationConfigurationHelper.__other_default_yaml_configuration_file_extension
            )
            LOGGER.debug(
                "Attempting to find/load '%s' as a default YAML configuration file.",
                new_file_name,
            )
            (
                did_apply_map,
                did_have_one_error,
            ) = ApplicationPropertiesYamlLoader.load_and_set(
                application_properties,
                new_file_name,
                handle_error_fn=handle_error_fn,
                clear_property_map=False,
                check_for_file_presence=True,
            )
        if not did_apply_map:
            LOGGER.debug("No default configuration files were loaded.")


# pylint: enable=too-few-public-methods
